#include "rwlock.h"



int
init_rwlock(rwlock_t *p)
{
	INITLOCK(p);
	sem_init(&p->readwait, 0, 0);
	sem_init(&p->writewait, 0, 0);
	p->nreaders = p->nwriters = 0;
	p->waitingreaders = p->waitingwriters = 0;
	return 0;
}

int
read_lock(volatile rwlock_t *l)
{
	LOCK(l);
	if (l->nwriters || l->waitingreaders || l->waitingwriters) {
		l->waitingreaders++;
		UNLOCK(l);
		sem_wait(&l->readwait);
		return 0;
	}
	l->nreaders++;
	UNLOCK(l);
	return 0;
}

int
read_unlock(volatile rwlock_t *rw)
{
	LOCK(rw);
	rw->nreaders--;
	if (rw->waitingwriters && rw->nreaders == 0) {
		rw->nwriters = 1;
		rw->waitingwriters--;
		UNLOCK(rw);
		sem_post(&rw->writewait);
		return 0;
	}
	UNLOCK(rw);
	return 0;
}


int
write_lock(volatile rwlock_t *rw)
{
	LOCK(rw);
	if (rw->nwriters || rw->nreaders) {
		rw->waitingwriters++;
		UNLOCK(rw);
		sem_wait(&rw->writewait);
		return 0;
	}
	rw->nwriters = 1;
	UNLOCK(rw);
	return 0;
}


int
write_unlock(volatile rwlock_t *rw)
{
	int      n;
	LOCK(rw);
/*-
 * Give priority to readers.
 */
	if (n=rw->waitingreaders) {
		rw->nwriters = 0;
		rw->nreaders = n;
		rw->waitingreaders = 0;
		UNLOCK(rw);
		while (n--)
			sem_post(&rw->readwait);
		return 0;
	}
/*-
 * none, wakeup one writer
 */
 	if (rw->waitingwriters) {
 		rw->waitingwriters--;
 		UNLOCK(rw);
 		sem_post(&rw->writewait);
 		return 0;
 	}
/*-
 * idle.
 */
	rw->nwriters = 0;
	UNLOCK(rw);
	return 0;
}
